/**
  *  \file t_fixed_list.cc
  */

#include "cpluslib/fixed_list.h"
#include "tests.h"
#include "cpluslib/string.h"
#include "cpluslib/format.h"

namespace {
    struct Counter {
        int n;
        static int c;

        Counter()
            : n(-9999)
            { ++c; }
        Counter(int n)
            : n(n)
            { ++c; }
        Counter(const Counter& o)
            : n(o.n)
            { ++c; }
        ~Counter()
            { --c; /*std::cout << "destroy " << n << "\n";*/ }

        bool operator==(const Counter& rhs) const
            { return n == rhs.n; }
        bool operator<(const Counter& rhs) const
            { return n < rhs.n; }
    };
}

int Counter::c;

static string_t
listToString(fixed_list<Counter>& v)
{
    string_t result = "{";
    for (fixed_list<Counter>::iterator i = v.begin(); i != v.end(); ++i) {
        if (i == v.begin())
            result += " ";
        else
            result += ", ";
        result += format("%d", i->n);
    }
    result += " }";
    return result;
}

template class heap_list<Counter>;

void
TestFixedList::testFixedList()
{
    typedef Counter T;

    /* Blank list */
    heap_list<T> is(20);
    TS_ASSERT_EQUALS(is.size(), 0);
    TS_ASSERT_EQUALS(listToString(is), "{ }");
    TS_ASSERT_EQUALS(Counter::c, 0);

    /* push_back */
    is.push_back(1);
    is.push_back(2);
    is.push_back(3);
    is.push_back(4);
    TS_ASSERT_EQUALS(is.size(), 4);
    TS_ASSERT_EQUALS(is.front(), 1);
    TS_ASSERT_EQUALS(is.back(), 4);
    TS_ASSERT_EQUALS(listToString(is), "{ 1, 2, 3, 4 }");
    TS_ASSERT_EQUALS(Counter::c, 4);

    /* pop_back */
    is.pop_back();
    TS_ASSERT_EQUALS(is.size(), 3);
    TS_ASSERT_EQUALS(is.front(), 1);
    TS_ASSERT_EQUALS(is.back(), 3);
    TS_ASSERT_EQUALS(listToString(is), "{ 1, 2, 3 }");
    TS_ASSERT_EQUALS(Counter::c, 3);

    /* pop_front */
    is.pop_front();
    TS_ASSERT_EQUALS(is.size(), 2);
    TS_ASSERT_EQUALS(is.front(), 2);
    TS_ASSERT_EQUALS(is.back(), 3);
    TS_ASSERT_EQUALS(listToString(is), "{ 2, 3 }");
    TS_ASSERT_EQUALS(Counter::c, 2);

    /* push_back again */
    is.push_back(4);
    is.push_back(5);
    is.push_back(6);
    is.push_back(7);
    const_cast<const heap_list<Counter>&>(is).rend();

    TS_ASSERT_EQUALS(is.size(), 6);
    TS_ASSERT_EQUALS(is.front(), 2);
    TS_ASSERT_EQUALS(is.back(), 7);
    TS_ASSERT_EQUALS(listToString(is), "{ 2, 3, 4, 5, 6, 7 }");
    TS_ASSERT_EQUALS(Counter::c, 6);

    /* Iteration */
    heap_list<T>::iterator i = is.begin(), j;
    TS_ASSERT_EQUALS(i->n, 2);
    ++i;
    TS_ASSERT_EQUALS(i->n, 3);
    ++i;
    TS_ASSERT_EQUALS(i->n, 4);

    /* erase */
    i = is.erase(i);
    TS_ASSERT_EQUALS(i->n, 5);
    TS_ASSERT_EQUALS(is.size(), 5);
    TS_ASSERT_EQUALS(listToString(is), "{ 2, 3, 5, 6, 7 }");
    TS_ASSERT_EQUALS(Counter::c, 5);

    /* insert */
    j = is.insert(i, 77);
    TS_ASSERT_EQUALS(is.size(), 6);
    TS_ASSERT_EQUALS(listToString(is), "{ 2, 3, 77, 5, 6, 7 }");
    TS_ASSERT_EQUALS(Counter::c, 6);
    TS_ASSERT_EQUALS(i->n, 5);
    TS_ASSERT_EQUALS(j->n, 77);

    /* push_front */
    is.push_front(333);
    TS_ASSERT_EQUALS(is.size(), 7);
    TS_ASSERT_EQUALS(listToString(is), "{ 333, 2, 3, 77, 5, 6, 7 }");
    TS_ASSERT_EQUALS(Counter::c, 7);

    /* insert front/end */
    is.insert(is.begin(), 332);
    TS_ASSERT_EQUALS(is.size(), 8);
    TS_ASSERT_EQUALS(listToString(is), "{ 332, 333, 2, 3, 77, 5, 6, 7 }");
    TS_ASSERT_EQUALS(Counter::c, 8);

    is.insert(is.end(), 999);
    TS_ASSERT_EQUALS(is.size(), 9);
    TS_ASSERT_EQUALS(listToString(is), "{ 332, 333, 2, 3, 77, 5, 6, 7, 999 }");
    TS_ASSERT_EQUALS(Counter::c, 9);

    /* reverse */
    is.reverse();
    TS_ASSERT_EQUALS(is.size(), 9);
    TS_ASSERT_EQUALS(listToString(is), "{ 999, 7, 6, 5, 77, 3, 2, 333, 332 }");
    TS_ASSERT_EQUALS(Counter::c, 9);

    /* iterators remain valid */
    TS_ASSERT_EQUALS(i->n, 5);
    TS_ASSERT_EQUALS(j->n, 77);

    /* remove by element */
    is.remove(77);
    TS_ASSERT_EQUALS(is.size(), 8);
    TS_ASSERT_EQUALS(listToString(is), "{ 999, 7, 6, 5, 3, 2, 333, 332 }");
    TS_ASSERT_EQUALS(Counter::c, 8);

    /* push_back again */
    is.push_back(332);
    is.push_back(332);
    is.push_back(331);
    TS_ASSERT_EQUALS(is.size(), 11);
    TS_ASSERT_EQUALS(listToString(is), "{ 999, 7, 6, 5, 3, 2, 333, 332, 332, 332, 331 }");
    TS_ASSERT_EQUALS(Counter::c, 11);

    /* unique */
    is.unique();
    TS_ASSERT_EQUALS(is.size(), 9);
    TS_ASSERT_EQUALS(listToString(is), "{ 999, 7, 6, 5, 3, 2, 333, 332, 331 }");
    TS_ASSERT_EQUALS(Counter::c, 9);

    /* sort */
    is.sort();
    TS_ASSERT_EQUALS(is.size(), 9);
    TS_ASSERT_EQUALS(listToString(is), "{ 2, 3, 5, 6, 7, 331, 332, 333, 999 }");
    TS_ASSERT_EQUALS(Counter::c, 9);

    /* static_list */
    static_list<T,10> ol;
    ol.push_back(1);
    ol.push_back(5);
    ol.push_back(11);
    ol.push_back(333);
    TS_ASSERT_EQUALS(ol.size(), 4);
    TS_ASSERT_EQUALS(listToString(ol), "{ 1, 5, 11, 333 }");
    TS_ASSERT_EQUALS(Counter::c, 13);

    /* merge */
    is.merge(ol);
    TS_ASSERT_EQUALS(is.size(), 13);
    TS_ASSERT_EQUALS(listToString(is), "{ 1, 2, 3, 5, 5, 6, 7, 11, 331, 332, 333, 333, 999 }");

    TS_ASSERT_EQUALS(ol.size(), 0);
    TS_ASSERT_EQUALS(listToString(ol), "{ }");
    TS_ASSERT_EQUALS(Counter::c, 13);

    /* clear is */
    static_list<T,20>().swap(is);
    TS_ASSERT_EQUALS(is.size(), 0);
    TS_ASSERT_EQUALS(listToString(is), "{ }");
    TS_ASSERT_EQUALS(Counter::c, 0);
}
